package tk.glover.jsfunittest.model;

import java.io.Serializable;
import java.util.Comparator;
import java.util.Date;

import javax.persistence.Column;
import javax.persistence.Embedded;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Table;

import org.hibernate.util.SerializationHelper;


/**
 * Address entity.
 * 
 * @author andrew glover
 */
@Entity
@Table(name = "jsfUnitTestAddresses")
public class Address implements Serializable, Comparable<Address> {

    public static final Comparator<Address> SORTER = new Comparator<Address>() {
        private final int FIELD_COUNT = 23;

        @Override
        public int compare(Address arg0, Address arg1) {
            int result = 0;
            for (int i = 0; i < FIELD_COUNT && result == 0; i++) {
                switch (i) {
                case 8:
                    result = compareStringsAsNumeric(arg0.addressSub.getStreetNumber(), arg1.addressSub.getStreetNumber());
                    break;
                case 9:
                    result = compareStrings(arg0.addressSub.getStreetName(), arg1.addressSub.getStreetName());
                    break;
                case 10:
                    result = compareStrings(arg0.addressSub.getStreetType(), arg1.addressSub.getStreetType());
                    break;
                case 13:
                    result = compareStrings(arg0.addressSub.getTown(), arg1.addressSub.getTown());
                    break;
                case 14:
                    result = compareStrings(arg0.addressSub.getPostcode(), arg1.addressSub.getPostcode());
                    break;
                case 15:
                    result = compareStrings(arg0.addressSub.getState(), arg1.addressSub.getState());
                    break;
                case 16:
                    result = compareStrings(arg0.addressSub.getCountry(), arg1.addressSub.getCountry());
                    break;
                case 18:
                    result = compareLongs(arg0.id, arg1.id);
                    break;
                case 22:
                    result = compareDates(arg0.addressSub.getStartDate(), arg1.addressSub.getStartDate());
                }
            }

            return result;
        }

        /**
         * Sorting not-null before null.
         * 
         * @param arg0
         *            an Object, potentially null
         * @param arg1
         *            another Object, potentially null
         * @return null if further testing is needed because none of the arguments are neither null or identical, 0 if both arguments are null or they
         *         are equal, 1 if arg0 is null, -1 if arg1 is null
         */
        private Integer compareObjects(Object arg0, Object arg1) {
            if (arg0 == null && arg1 == null) {
                return Integer.valueOf(0);
            }

            if (arg0 != null && arg1 != null) {
                return arg0 == arg1 ? Integer.valueOf(0) : null;
            }

            return Integer.valueOf(arg0 == null ? 1 : -1);
        }

        /**
         * Sorting not-empty before empty. String#compareTo() would order it the other way.
         * 
         * Should only be invoked after compareObjects, when that methods returns null; or else NPE.
         * 
         * @param arg0
         *            a non-null String, potentially empty
         * @param arg1
         *            another non-null String, potentially empty
         * @return null if further testing is needed because none of the arguments are empty 0 if both arguments are empty 1 if arg0 is empty -1 if
         *         arg1 is empty
         */
        private Integer compareEmptyStrings(String arg0, String arg1) {
            int length0 = arg0.length();
            int length1 = arg1.length();

            if (length0 == 0 && length1 == 0) {
                return Integer.valueOf(0);
            }

            if (length0 > 0 && length1 > 0) {
                return null;
            }

            return Integer.valueOf(length0 == 0 ? 1 : -1);
        }

        private int compareStrings(String arg0, String arg1) {
            Integer comparedObjects = compareObjects(arg0, arg1);
            if (comparedObjects != null) {
                return comparedObjects.intValue();
            }

            Integer comparedEmpty = compareEmptyStrings(arg0, arg1);
            if (comparedEmpty != null) {
                return comparedEmpty.intValue();
            }

            return arg0.compareTo(arg1);
        }

        private int compareLongs(Long arg0, Long arg1) {
            Integer comparedObjects = compareObjects(arg0, arg1);

            return comparedObjects == null ? arg0.compareTo(arg1) : comparedObjects.intValue();
        }

        private int compareDates(Date arg0, Date arg1) {
            Integer comparedObjects = compareObjects(arg0, arg1);

            return comparedObjects == null ? arg0.compareTo(arg1) : comparedObjects.intValue();
        }

        private int compareStringsAsNumeric(String arg0, String arg1) {
            Integer comparedObjects = compareObjects(arg0, arg1);

            if (comparedObjects != null) {
                return comparedObjects.intValue();
            }

            Integer comparedEmpty = compareEmptyStrings(arg0, arg1);

            if (comparedEmpty != null) {
                return comparedEmpty.intValue();
            }

            Integer numricArg0 = extractNumber(arg0);
            Integer numricArg1 = extractNumber(arg1);

            if (numricArg0 != null && numricArg1 != null) {
                return numricArg0.compareTo(numricArg1);
            }

            return arg0.compareTo(arg1);
        }

        /**
         * 
         * @param potentialNumber
         *            , must not be null
         * @return the integer value of potentialNumber if it is convertible, or else if potentialNumber starts with a sequence of digits only, the
         *         integer value of that sequence, or else null
         */
        private Integer extractNumber(String potentialNumber) {
            Integer result = null;

            String trimmed = potentialNumber.trim();

            try {
                result = Integer.valueOf(trimmed);
            } catch (NumberFormatException e) {
                int i = 0;
                int length = trimmed.length();

                for (; i < length; i++) {
                    if (!Character.isDigit(trimmed.charAt(i))) {
                        break;
                    }
                }

                if (i > 0) {
                    result = Integer.valueOf(trimmed.substring(0, i));
                }
            }

            return result;
        }
    };

    /** The Constant serialVersionUID. */
    private static final long serialVersionUID = -6573571502688001174L;

    /** The id. */
    @Id
    @GeneratedValue(strategy = GenerationType.AUTO)
    @Column(name = "ADDR_ID")
    private Long id;


    /** The address sub. */
    @Embedded
    protected AddressSub addressSub = new AddressSub();


    /**
     * Instantiates a new address.
     */
    public Address() {
        super();
        addressSub = new AddressSub();
    }

    /**
     * Checks if is transient.
     * 
     * @return true, if is transient
     */
    public boolean isTransient() {
        return this.id == null;
    }

    /**
     * Checks if is bad address.
     * 
     * @return true, if is bad address
     */
    public boolean isBadAddress() {
        return this.addressSub.getStreetNumber() == null || this.addressSub.getStreetName() == null || this.addressSub.getTown() == null
                        || this.addressSub.getPostcode() == null || this.addressSub.getState() == null;
    }

    /**
     * Gets the id.
     * 
     * @return the id
     */
    public Long getId() {
        return id;
    }

    /**
     * Sets the id.
     * 
     * @param id
     *            the new id
     */
    public void setId(Long id) {
        this.id = id;
    }

    /**
     * Gets the address sub.
     * 
     * @return the address sub
     */
    public AddressSub getAddressSub() {
        return addressSub;
    }

    /**
     * Sets the address sub.
     * 
     * @param addressSub
     *            the new address sub
     */
    public void setAddressSub(AddressSub addressSub) {
        this.addressSub = addressSub;
    }



    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        // following code based on RG200_LOAD_ADDRESS_DETAILS
        StringBuffer formatedAddress = new StringBuffer();
        formatedAddress.append(addressSub.toString());
        // white space removal added to handle MARS-365 (extra white space in default address)
        return formatedAddress.toString().trim().replaceAll("\\s+", " ");
    }


    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((addressSub == null) ? 0 : addressSub.hashCode());
        result = prime * result + ((id == null) ? 0 : id.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        Address other = (Address) obj;
        if (addressSub == null) {
            if (other.addressSub != null)
                return false;
        } else if (!addressSub.equals(other.addressSub))
            return false;
        if (id == null) {
            if (other.id != null)
                return false;
        } else if (!id.equals(other.id))
            return false;
        return true;
    }

    @Override
    public int compareTo(Address o) {
        return this.toString().compareTo(o.toString());
    }

    /**
     * Factory method to create a new address from passing address. The new address only has the same attributes as the given address, but not
     * relationships.
     * 
     * @param newAddress
     *            the new address
     * 
     * @return the address
     */
    public void copyAttributesFrom(Address newAddress) {
        this.addressSub = (AddressSub) SerializationHelper.clone(newAddress.getAddressSub());
    }

}